from typing import List, Dict, Any

import numpy as np
from rlgym.api import RewardFunction, AgentID
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import BACK_WALL_Y

RAMP_HEIGHT = 256


class AerialDistanceReward(RewardFunction[AgentID, GameState, float]):
    """
    Aerial distance reward.
    - First aerial touch is rewarded by height
    - Consecutive touches based on distance travelled (since last aerial touch)
    - Resets when grounded or when another player touches the ball
    """

    def __init__(
            self,
            touch_height_weight: float = 1.0,
            car_distance_weight: float = 1.0,
            ball_distance_weight: float = 1.0,
            distance_normalization: float = 1 / BACK_WALL_Y
    ):
        super().__init__()
        self.touch_height_weight = touch_height_weight
        self.car_distance_weight = car_distance_weight
        self.ball_distance_weight = ball_distance_weight
        self.distance_normalization = distance_normalization
        self.distances = {}
        self.last_touch_agent = None
        self.prev_state = None

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.distances = {k: 0 for k in agents}
        self.last_touch_agent = None
        self.prev_state = initial_state

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        rewards = {k: 0 for k in agents}
        for agent in agents:
            car = state.cars[agent]
            if self.last_touch_agent == agent:
                if car.physics.position[2] < RAMP_HEIGHT:
                    self.distances[agent] = 0
                    self.last_touch_agent = None
                else:
                    dist_car = np.linalg.norm(car.physics.position - self.prev_state.cars[agent].physics.position)
                    dist_ball = np.linalg.norm(state.ball.position - self.prev_state.ball.position)
                    self.distances[agent] += (dist_car * self.car_distance_weight
                                              + dist_ball * self.ball_distance_weight)
            if car.ball_touches > 0:
                if self.last_touch_agent == agent:
                    norm_dist = self.distances[agent] * self.distance_normalization
                    rewards[agent] += norm_dist
                else:
                    w1 = self.car_distance_weight
                    w2 = self.ball_distance_weight
                    if w1 == w2 == 0:
                        w1 = w2 = 1
                    touch_height = float((w1 * car.physics.position[2] + w2 * state.ball.position[2]) / (w1 + w2))
                    touch_height = max(0.0, touch_height - RAMP_HEIGHT)  # Clamp to 0
                    norm_dist = touch_height * self.distance_normalization
                    rewards[agent] += norm_dist * self.touch_height_weight
                    self.last_touch_agent = agent
                self.distances[agent] = 0
        self.prev_state = state
        shared_info["aerial_distance_info"] = {"distances": self.distances, "last_touch_agent": self.last_touch_agent}
        return rewards



from typing import List, Dict, Any

from rlgym.api import RewardFunction, AgentID, StateType, RewardType
from rlgym.rocket_league.api import GameState


class EpisodeEndReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, reward: float = 1.0):
        self.reward = reward

    def reset(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> None:
        pass

    def get_rewards(self, agents: List[AgentID], state: StateType, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, RewardType]:
        return {
            a: self.reward if is_terminated[a] else 0.0
            for a in agents
        }



import numpy as np
from rlgym.rocket_league.common_values import CAR_MAX_ANG_VEL, TICKS_PER_SECOND

# Code by the legendary Sam Mish, from https://www.smish.dev/rocket_league/inverse_aerial_control/
# NOTE: Modified to produce full-magnitude inputs when at max angvel (thanks Zealan)


T_r = -36.07956616966136  # torque coefficient for roll
T_p = -12.14599781908070  # torque coefficient for pitch
T_y = 8.91962804287785  # torque coefficient for yaw
D_r = -4.47166302201591  # drag coefficient for roll
D_p = -2.798194258050845  # drag coefficient for pitch
D_y = -1.886491900437232  # drag coefficient for yaw


def aerial_inputs(ang_vel_start, ang_vel_end, rot_mat_start, rot_mat_end, dt, is_flipping=False):
    scale = 1.0
    if np.linalg.norm(ang_vel_end) >= CAR_MAX_ANG_VEL - 0.01:
        scale = 1.25  # Scale up so we don't get partial inputs when we hit the max angular velocity
    tau = (ang_vel_end * scale - ang_vel_start) / dt  # net torque in world coordinates
    tst = np.transpose(rot_mat_start)
    # tau1 = np.dot(tst, tau)  # net torque in local coordinates
    tau = np.array([
        tst[0, 0] * tau[0] + tst[0, 1] * tau[1] + tst[0, 2] * tau[2],
        tst[1, 0] * tau[0] + tst[1, 1] * tau[1] + tst[1, 2] * tau[2],
        tst[2, 0] * tau[0] + tst[2, 1] * tau[1] + tst[2, 2] * tau[2]
    ])
    # omega_local1 = np.dot(tst, omega_start)  # beginning-step angular velocity in local coordinates
    ang_vel_local_start = np.array([
        tst[0, 0] * ang_vel_start[0] + tst[0, 1] * ang_vel_start[1] + tst[0, 2] * ang_vel_start[2],
        tst[1, 0] * ang_vel_start[0] + tst[1, 1] * ang_vel_start[1] + tst[1, 2] * ang_vel_start[2],
        tst[2, 0] * ang_vel_start[0] + tst[2, 1] * ang_vel_start[1] + tst[2, 2] * ang_vel_start[2]
    ])

    # assert np.allclose(tau1, tau, equal_nan=True)
    # assert np.allclose(omega_local1, ang_vel_local_start, equal_nan=True)

    rhs = np.array([
        tau[0] - D_r * ang_vel_local_start[0],
        tau[1] - D_p * ang_vel_local_start[1],
        tau[2] - D_y * ang_vel_local_start[2]
    ])

    u = np.array([
        rhs[0] / T_r,  # roll
        rhs[1] / (T_p + np.sign(rhs[1]) * ang_vel_local_start[1] * D_p),  # pitch
        rhs[2] / (T_y - np.sign(rhs[2]) * ang_vel_local_start[2] * D_y)  # yaw
    ])

    # ensure values are between -1 and +1
    u = np.clip(u, -1, +1)

    if is_flipping:
        # From https://github.com/ZealanL/RLCarInputSolver/blob/main/src/AirSolver.cpp
        tst = np.transpose(rot_mat_end)
        ang_vel_local_end = np.array([
            tst[0, 0] * ang_vel_end[0] + tst[0, 1] * ang_vel_end[1] + tst[0, 2] * ang_vel_end[2],
            tst[1, 0] * ang_vel_end[0] + tst[1, 1] * ang_vel_end[1] + tst[1, 2] * ang_vel_end[2],
            tst[2, 0] * ang_vel_end[0] + tst[2, 1] * ang_vel_end[1] + tst[2, 2] * ang_vel_end[2]
        ])

        if abs(ang_vel_local_start[1]) > abs(ang_vel_local_end[1]):
            # Flip cancel
            u[1] = np.sign(ang_vel_local_start[1])

    return u[1], u[2], u[0]  # pitch, yaw, roll



from typing import Any, Dict, List

import numpy as np
from rlgym.api import RewardFunction, AgentID
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import BALL_MAX_SPEED


class AdvancedTouchReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, touch_reward: float = 1.0, acceleration_reward: float = 0.0, use_touch_count: bool = True):
        self.touch_reward = touch_reward
        self.acceleration_reward = acceleration_reward
        self.use_touch_count = use_touch_count

        self.prev_ball = None

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.prev_ball = initial_state.ball

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        rewards = {agent: 0 for agent in agents}
        ball = state.ball
        for agent in agents:
            touches = state.cars[agent].ball_touches

            if touches > 0:
                if not self.use_touch_count:
                    touches = 1
                acceleration = np.linalg.norm(ball.linear_velocity - self.prev_ball.linear_velocity) / BALL_MAX_SPEED
                rewards[agent] += self.touch_reward * touches
                rewards[agent] += acceleration * self.acceleration_reward

        self.prev_ball = ball

        return rewards



from typing import List, Dict, Any

from rlgym.api import RewardFunction, AgentID
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import BLUE_TEAM

from rlgym_tools.rocket_league.math.ball import GOAL_THRESHOLD
from rlgym_tools.rocket_league.math.solid_angle import view_goal_ratio


class GoalProbReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, gamma: float = 1):
        """
        According to Ng. et al. (1999), a reward shaping function must be of the form:
        F(s, a, s') = γ * Φ(s') - Φ(s)
        to preserve all the optimal policies of the original MDP,
        where Φ(s) is a function that estimates the potential of a state.
        The gamma term is supposed to be the same as the one used to discount future rewards.
        Here it serves to adjust for the fact that it will be discounted in the future.
        In practice though, leaving it as 1 is probably fine.
        (in fact the paper only deals with finite MDPs with γ=1 and infinite MDPs with γ<1,
        whereas we typically have a finite MDP with γ<1)

        :param gamma: the discount factor for the reward shaping function.
        """
        self.prob = None
        self.gamma = gamma

    def calculate_blue_goal_prob(self, state: GameState):
        """
        Calculate the probability of a goal being scored *by blue*, e.g. on the orange goal, from the current state.

        :param state: the current game state
        :return: the probability of a goal being scored by blue
        """
        raise NotImplementedError

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.prob = self.calculate_blue_goal_prob(initial_state)

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        if state.goal_scored:
            if state.scoring_team == BLUE_TEAM:
                prob = 1
            else:
                prob = 0
        else:
            prob = self.calculate_blue_goal_prob(state)
        # Probability goes from 0-1, but for a reward we want it to go from -1 to 1
        # 2x-1 - (2y-1) = 2(x-y)
        reward = 2 * (self.gamma * prob - self.prob)
        rewards = {
            agent: reward if state.cars[agent].is_blue else -reward
            for agent in agents
        }
        self.prob = prob
        return rewards


class GoalViewReward(GoalProbReward):
    """
    Simple estimate based on the apparent size of each goal.
    Basically it says "if we cast a ray from the ball in random directions until it hits a goal,
    what's the chance it hits the orange goal (blue scoring)?"
    """

    def calculate_blue_goal_prob(self, state: GameState):
        ball_pos = state.ball.position
        view_blue = view_goal_ratio(ball_pos, -GOAL_THRESHOLD)  # Blue net aka orange scoring
        view_orange = view_goal_ratio(ball_pos, GOAL_THRESHOLD)  # Orange net aka blue scoring
        return view_orange / (view_blue + view_orange)



from typing import Dict, Any

import numpy as np
from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import SIDE_WALL_X, BACK_WALL_Y, CEILING_Z
from rlgym.rocket_league.math import rand_vec3, rand_uvec3


class RandomPhysicsMutator(StateMutator[GameState]):
    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        padding = 100
        for po in [state.ball] + [car.physics for car in state.cars.values()]:
            po.position = np.random.uniform([-SIDE_WALL_X + padding, -BACK_WALL_Y + padding, 0 + padding],
                                            [SIDE_WALL_X - padding, BACK_WALL_Y - padding, CEILING_Z - padding])
            po.linear_velocity = rand_vec3(2300)
            po.angular_velocity = rand_vec3(5)
            fw = rand_uvec3()
            up = rand_uvec3()
            right = np.cross(fw, up)
            po.rotation_mtx = np.stack([fw, right, up])



from typing import Dict, Any, List

import numpy as np
from rlgym.api import TransitionEngine
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.sim import RocketSimEngine


class MultiEnvEngine(TransitionEngine[tuple, List[GameState], np.ndarray]):
    """
    MultiEnvEngine is a wrapper around multiple RocketSimEngine instances. It allows for multiple environments to be
    run in "parallel". While the environments are not actually run in parallel, it can still be useful since performing
    batched model inference is usually faster than performing inference on a single instance at a time.
    """

    def __init__(self, num_envs: int):
        # TODO use the parallel step method from RocketSim bindings
        self.envs = [
            RocketSimEngine()
            for _ in range(num_envs)
        ]

    @property
    def agents(self) -> List[tuple]:
        agents = []
        for i, env in enumerate(self.envs):
            for agent in env.agents:
                agents.append((i, agent))
        return agents

    @property
    def max_num_agents(self) -> int:
        return sum(env.max_num_agents for env in self.envs)

    @property
    def state(self) -> List[GameState]:
        return [env.state for env in self.envs]

    @property
    def config(self) -> Dict[str, Any]:
        configs = [env.config for env in self.envs]
        keys = configs[0].keys()
        return {key: [config[key] for config in configs] for key in keys}

    @config.setter
    def config(self, value: Dict[str, Any]):
        for i in range(len(self.envs)):
            # Make sure we use RocketSimEngine's config.setter
            self.envs[i].config = {  # noqa setter not implemented
                key: value[key][i]
                for key in value.keys()
            }

    def step(self, actions: Dict[tuple, np.ndarray], shared_info: Dict[str, Any]) -> List[GameState]:
        grouped_actions = [{} for _ in range(len(self.envs))]
        for (i, agent), action in actions.items():
            grouped_actions[i][agent] = action
        states = []
        for i, env in enumerate(self.envs):
            state = env.step(grouped_actions[i], shared_info)
            states.append(state)
        return states

    def create_base_state(self) -> List[GameState]:
        return [env.create_base_state() for env in self.envs]

    def set_state(self, desired_state: List[GameState], shared_info: Dict[str, Any]) -> List[GameState]:
        for i, env in enumerate(self.envs):
            env.set_state(desired_state[i], shared_info)
        return self.state

    def close(self) -> None:
        for env in self.envs:
            env.close()



from typing import Tuple, Dict, Any, List

import numpy as np
from rlgym.api import ActionParser, AgentID, ActionType, StateType, EngineActionType, ActionSpaceType
from rlgym.rocket_league.api import GameState


class DelayedAction(ActionParser[AgentID, np.ndarray, np.ndarray, GameState, Tuple[AgentID, np.ndarray]]):
    """
    DelayedAction delays all actions by a specified number of ticks. The last tick(s) of the previous action are
    prepended to the current action. RLBot has a 1-tick delay so this is essential for full reproducibility.
    """

    def __init__(self, action_parser: ActionParser, delay_ticks: int = 1):
        self.action_parser = action_parser
        self.delay_ticks = delay_ticks
        self.delayed_actions = None

    def get_action_space(self, agent: AgentID) -> ActionSpaceType:
        return self.action_parser.get_action_space(agent)

    def reset(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> None:
        self.action_parser.reset(agents, initial_state, shared_info)
        self.delayed_actions = {agent: np.zeros((self.delay_ticks, 8)) for agent in agents}

    def parse_actions(self, actions: Dict[AgentID, ActionType], state: StateType, shared_info: Dict[str, Any]) \
            -> Dict[AgentID, EngineActionType]:
        parsed_actions = self.action_parser.parse_actions(actions, state, shared_info)
        returned_actions = {}
        for agent, action in parsed_actions.items():
            del_action = self.delayed_actions[agent]
            ret_action = np.zeros_like(action)
            ret_action[:self.delay_ticks] = del_action
            ret_action[self.delay_ticks:] = action[:-self.delay_ticks]
            del_action[:] = action[-self.delay_ticks:]
            returned_actions[agent] = ret_action
        shared_info["delayed_actions"] = self.delayed_actions
        return returned_actions



import math

import numpy as np
from rlgym.rocket_league.common_values import BACK_WALL_Y, SIDE_WALL_X, GOAL_HEIGHT

try:
    from scipy.stats.distributions import skellam
except ImportError:
    skellam = None  # Lets us import the constants without scipy. It'll fail if you try to use the function.

# Win probability estimation using Skellam distribution

AVERAGE_EPISODE_LENGTH = np.array([26.32, 47.16, 68.52])  # In seconds
GOALS_PER_MINUTE = 60 / AVERAGE_EPISODE_LENGTH

FLOOR_AREA = 4 * BACK_WALL_Y * SIDE_WALL_X - 1152 * 1152  # Subtract corners
GOAL_AREA = GOAL_HEIGHT * 880
BASE_FLOOR_PROB = FLOOR_AREA / (2 * GOAL_AREA + FLOOR_AREA)
SECONDS_PER_MINUTE = 60


def win_prob(goals_per_minute: float, time_left_seconds: float, differential: int,
             next_goal_prob: float = 0.5, hit_ground_prob: float = BASE_FLOOR_PROB):
    if skellam is None:
        raise ImportError("scipy is required to use the win_prob function")
    # First, make it symmetric, so we only need to concern ourselves with non-negative differentials
    if differential < 0:
        p_loss = win_prob(goals_per_minute, time_left_seconds, -differential,
                          1 - next_goal_prob, hit_ground_prob)
        return 1 - p_loss

    if math.isinf(time_left_seconds):
        # Overtime, next goal wins
        return next_goal_prob
    if differential == 0 and next_goal_prob == 0.5:
        return 0.5

    # We win if we score, or go to OT and win there (50%)
    win_prob_if_tied = (1 - hit_ground_prob) * next_goal_prob + 0.5 * hit_ground_prob
    # We win if ball hits the ground or we score
    win_prob_if_one_ahead = hit_ground_prob + (1 - hit_ground_prob) * next_goal_prob
    # We win only by scoring without it hitting the ground
    win_prob_if_one_behind = (1 - hit_ground_prob) * next_goal_prob

    p_two_goal_lead = p_one_goal_lead = p_tied = p_one_goal_deficit = 0
    if time_left_seconds <= 0:
        if differential == 0:
            p_tied = 1
        elif differential == 1:
            p_one_goal_lead = 1
        else:  # differential >= 2:
            p_two_goal_lead = 1
    else:
        rate = goals_per_minute * time_left_seconds / SECONDS_PER_MINUTE

        dist = skellam(rate / 2, rate / 2)
        p_two_goal_lead = dist.cdf(differential - 2)
        p_one_goal_lead, p_tied, p_one_goal_deficit = dist.pmf([differential - 1, differential, differential + 1])

    p_win = (
            p_two_goal_lead  # No way for opponent to win
            + p_one_goal_lead * win_prob_if_one_ahead
            + p_tied * win_prob_if_tied
            + p_one_goal_deficit * win_prob_if_one_behind
    )

    return p_win



import math

import numpy as np
from rlgym.rocket_league.common_values import BALL_RADIUS, GOAL_HEIGHT, GOAL_CENTER_TO_POST


def solid_angle_eriksson(O, A, B, C):
    # Calculate the solid angle of a triangle ABC from the point O
    a = A - O
    b = B - O
    c = C - O
    a /= np.linalg.norm(a)
    b /= np.linalg.norm(b)
    c /= np.linalg.norm(c)
    numerator = np.linalg.norm(np.dot(np.cross(a, b), c))  # noqa numpy complains about cross for no reason
    denominator = 1 + np.dot(a, b) + np.dot(b, c) + np.dot(c, a)
    E = 2 * math.atan2(numerator, denominator)
    return E


def view_goal_ratio(pos, goal_y, margin=BALL_RADIUS):
    # Calculate the percent of the field of view that the goal takes up
    max_x = GOAL_CENTER_TO_POST - margin
    min_x = -max_x
    max_z = GOAL_HEIGHT - margin
    min_z = margin
    bl = np.array([min_x, goal_y, min_z])
    br = np.array([max_x, goal_y, min_z])
    tl = np.array([min_x, goal_y, max_z])
    tr = np.array([max_x, goal_y, max_z])
    solid_angle_1 = solid_angle_eriksson(pos, bl, br, tl)
    solid_angle_2 = solid_angle_eriksson(pos, br, tr, tl)
    return (solid_angle_1 + solid_angle_2) / (4 * math.pi)


def solid_angle_ball(pos, ball_pos, ball_radius=BALL_RADIUS):
    # Calculate the solid angle of the ball
    d = np.linalg.norm(pos - ball_pos)
    r_sphere = math.sqrt(d ** 2 - ball_radius ** 2)
    E = 2 * math.pi * (1 - r_sphere / d)
    return E


def view_ball_ratio(pos, ball_pos):
    # Calculate the percent of the field of view that the ball takes up
    solid_angle = solid_angle_ball(pos, ball_pos)
    return solid_angle / (4 * math.pi)



from typing import Dict, Any, Tuple, List

import numpy as np
from rlgym.api import ActionParser, AgentID, ActionType, EngineActionType, ActionSpaceType
from rlgym.rocket_league.api import GameState


class QueuedAction(ActionParser[AgentID, np.ndarray, np.ndarray, GameState, Tuple[AgentID, np.ndarray]]):
    def __init__(self, parser: ActionParser, action_queue_size: int = 1):
        """
        QueuedAction maintains a queue of actions to execute and adds parsed actions to the queue.

        :param parser: the action parser to parse actions that are then added to the queue.
        """
        super().__init__()
        self.parser = parser
        self.action_queue_size = action_queue_size
        self.action_queue = {}
        self.is_initial = True

    def get_action_space(self, agent: AgentID) -> ActionSpaceType:
        return self.parser.get_action_space(agent)

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.parser.reset(agents, initial_state, shared_info)
        self.action_queue = {k: [] for k in initial_state.cars.keys()}
        self.is_initial = True
        shared_info["action_queue"] = self.action_queue

    def parse_actions(self, actions: Dict[AgentID, ActionType], state: GameState, shared_info: Dict[str, Any]) \
            -> Dict[AgentID, EngineActionType]:
        parsed_actions = self.parser.parse_actions(actions, state, shared_info)
        returned_actions = {}
        if self.is_initial:
            for agent, action in parsed_actions.items():
                self.action_queue[agent] = [action] * self.action_queue_size
            self.is_initial = False
        for agent, action in parsed_actions.items():
            self.action_queue[agent].append(action)
            returned_actions[agent] = self.action_queue[agent].pop(0)
        shared_info["action_queue"] = self.action_queue
        return returned_actions



from typing import Tuple, Dict, Any, List

import numpy as np
from rlgym.api import ActionParser, AgentID, ActionType, StateType, EngineActionType, ActionSpaceType
from rlgym.rocket_league.api import GameState


class ActionHistoryWrapper(ActionParser[AgentID, np.ndarray, np.ndarray, GameState, Tuple[AgentID, np.ndarray]]):
    def __init__(self, action_parser: ActionParser, ticks_to_remember: int = 8):
        super().__init__()
        self.action_parser = action_parser
        self.action_history = {}
        self.ticks_to_remember = ticks_to_remember

    def get_action_space(self, agent: AgentID) -> ActionSpaceType:
        return self.action_parser.get_action_space(agent)

    def reset(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> None:
        self.action_parser.reset(agents, initial_state, shared_info)
        self.action_history = {agent: np.zeros((0, 8)) for agent in agents}

    def parse_actions(self, actions: Dict[AgentID, ActionType], state: GameState, shared_info: Dict[str, Any]) \
            -> Dict[AgentID, EngineActionType]:
        parsed_actions = self.action_parser.parse_actions(actions, state, shared_info)
        for agent, action in parsed_actions.items():
            # Action has shape (ticks, 8)
            history = self.action_history[agent]
            if len(history) < self.ticks_to_remember:
                self.action_history[agent] = np.vstack((history, action))
            elif len(action) >= self.ticks_to_remember:
                self.action_history[agent] = action[-self.ticks_to_remember:]
            else:
                history[:-len(action)] = history[len(action):]
                history[-len(action):] = action
        shared_info["action_history"] = self.action_history
        return parsed_actions



import numpy as np
from rlgym.rocket_league.api import Car


def filter_action_options(car: Car, replay_action: np.ndarray, action_options: np.ndarray, greedy_continuous: bool,
                          deadzone: float = 0.5):
    # Filters out actions that are definitely not correct
    margin = 1e-3

    candidates = action_options.copy()

    jump_matters = (car.on_ground or car.can_flip) and not car.is_holding_jump
    boost_matters = car.boost_amount > 0
    is_grounded = car.on_ground and replay_action[5] == 0
    is_aerial = not car.on_ground and replay_action[5] == 0

    if jump_matters:
        # Jump
        jump_error = np.abs(candidates[:, 5] - replay_action[5])
        candidates = candidates[jump_error == jump_error.min()]

        if not car.on_ground and replay_action[5] == 1:
            # Check if it's a dodge or a double jump
            is_dodge = np.abs(replay_action[2:5]).sum() >= deadzone
            is_dodges = np.abs(candidates[:, 2:5]).sum(axis=1) >= deadzone  # Jump is already filtered
            dodge_error = np.abs(is_dodges - float(is_dodge))
            candidates = candidates[dodge_error == dodge_error.min()]

            if is_dodge:
                # Check that the dodge direction error is minimized
                dodge_dir = np.array([replay_action[2], replay_action[3] + replay_action[4]])
                dodge_dirs = np.array([candidates[:, 2], candidates[:, 3] + candidates[:, 4]]).T
                if np.all(dodge_dir == 0):
                    # Stall
                    dir_dot = np.all(dodge_dirs == 0, axis=1)
                else:
                    dodge_dir /= np.linalg.norm(dodge_dir)
                    m = np.any(dodge_dirs != 0, axis=1)
                    dodge_dirs[m] /= np.linalg.norm(dodge_dirs[m], axis=1, keepdims=True)
                    dir_dot = np.dot(dodge_dirs, dodge_dir)
                candidates = candidates[dir_dot == dir_dot.max()]

    if boost_matters:
        # Boost
        boost_error = np.abs(candidates[:, 6] - replay_action[6])
        candidates = candidates[boost_error == boost_error.min()]

    if is_grounded:  # Grounded
        # Handbrake
        handbrake_error = np.abs(candidates[:, 7] - replay_action[7])
        candidates = candidates[handbrake_error == handbrake_error.min()]

        # Throttle
        throttle = replay_action[0] if (replay_action[6] == 0 and boost_matters) else 1
        if abs(throttle) >= margin:
            # There is a braking effect when
            throttle_dir = np.sign(throttle)
            throttle_dirs = np.sign(candidates[:, 0]) * (abs(candidates[:, 0]) >= margin)
            dir_error = throttle_dirs == -throttle_dir
            candidates = candidates[dir_error == dir_error.min()]

        if greedy_continuous:
            throttle_error = np.abs(candidates[:, 0] - throttle)
            candidates = candidates[throttle_error == throttle_error.min()]

            steer_error = np.abs(candidates[:, 1] - replay_action[1])
            candidates = candidates[steer_error == steer_error.min()]
    elif is_aerial:  # Aerial
        if greedy_continuous:
            # Pitch, yaw, roll
            rotate_dir = replay_action[2:5]
            rotate_dirs = candidates[:, 2:5]

            dir_error = np.linalg.norm(rotate_dirs - rotate_dir, axis=1)
            candidates = candidates[dir_error == dir_error.min()]

    # Steer (on ground) and pitch/yaw/roll (in air) are ignored due to being continuous
    return candidates


def get_best_action_options(car: Car, replay_action: np.ndarray, action_options: np.ndarray, dodge_deadzone=0.5,
                            greedy=True) -> np.ndarray:
    """
    Produces a probability distribution over the action options, where the best options are given the highest
    probability.

    :param car: The car state
    :param replay_action: The action to match
    :param action_options: The available action options
    :param dodge_deadzone: The deadzone for dodges
    :param greedy: If True, after filtering candidates it picks the closest ones by Euclidean distance.
    :return: The probabilities for each action option
    """
    candidates = filter_action_options(car, replay_action, action_options, True, dodge_deadzone)

    if greedy:
        naive_error = np.linalg.norm(candidates - replay_action, axis=1)
        candidates = candidates[naive_error == naive_error.min()]

    probs = np.zeros(len(action_options), dtype=np.float32)
    for i, c in enumerate(candidates):
        probs[np.all(action_options == c, axis=1)] = 1
    probs /= probs.sum()

    return probs


def get_weighted_action_options(car: Car, replay_action: np.ndarray, action_options: np.ndarray, dodge_deadzone=0.5):
    """
    Produces weights for each action such that the weighted sum of the action options is as close as possible to
    replay_action, and that the sum of the weights is 1 (e.g. they represent probabilities).

    We want to combine only the two available values closest to the true value, e.g. if we can have 0.5 and 1,
    then 0.75 would be a 50/50 mix of the two.

    :param car: The car state
    :param replay_action: The action to match
    :param action_options: The available action options
    :param dodge_deadzone: The deadzone for dodges
    :return: The weights for each action option
    """
    # Mostly the same as get_best_action_option, but slightly more lenient with candidates,
    # so we can get closer with the weighted average
    margin = 0.01
    candidates = filter_action_options(car, replay_action, action_options, False, dodge_deadzone)

    button_weights = np.ones(8)

    def _validate_continuous(button):
        # Returns a boolean mask of valid candidates (e.g. between the bins in the action options)
        replay_val = replay_action[button]
        o = candidates[:, button]
        diff = np.abs(o - replay_val)
        if diff.min() <= margin:
            v = diff <= margin
        else:
            # Include the bin borders surrounding the replay value
            unique = np.unique(o)
            bins = np.digitize(o, unique)
            diff = bins - np.digitize(replay_val, unique)
            v = (diff == 0) | (diff == 1)
        return v

    if car.on_ground and replay_action[5] == 0:
        handbrake_error = np.abs(candidates[:, 7] - replay_action[7])
        candidates = candidates[handbrake_error == handbrake_error.min()]

        candidates = candidates[_validate_continuous(0)]  # Throttle
        candidates = candidates[_validate_continuous(1)]  # Steer
        button_weights[[2, 3, 4]] = 1e-2  # Basically telling it to only optimize this if it has options
        button_weights[[5, 6]] = 1e-4
    elif replay_action[5] == 0 or car.on_ground or not car.can_flip or car.is_holding_jump:
        button_weights[[0, 1, 7]] = 1e-2
        button_weights[[5, 6]] = 1e-4
        candidates = candidates[_validate_continuous(2)]  # Pitch
        candidates = candidates[_validate_continuous(3)]  # Yaw
        candidates = candidates[_validate_continuous(4)]  # Roll

        if len(candidates) > 1 and np.all(candidates[:, [2, 3, 4]] == candidates[0, [2, 3, 4]]):
            candidates = candidates[_validate_continuous(0)]  # Throttle

    if len(candidates) == 1:
        weights = np.zeros(len(action_options))
        idx = np.where(np.all(action_options == candidates[0], axis=1))[0]
        weights[idx] = 1
        return weights

    # Perform least squares with sum(weights) = 1 constraint
    scale = 1000  # Scaling factor to control the importance of the constraint
    coefs = np.concatenate([(candidates * button_weights).T, scale * np.ones((1, len(candidates)))])
    target = np.concatenate([(replay_action * button_weights), [scale]])

    last_weights = np.zeros(len(candidates))
    while True:
        weights = np.linalg.lstsq(coefs, target, rcond=None)[0]
        valid = weights >= -1e-5  # Numerical instability can cause very small negative weights
        if valid.all() or np.allclose(weights, last_weights):
            break
        coefs[:, ~valid] = 0  # Set the vectors to 0, so they are effectively removed
        last_weights = weights

    weights = weights / weights.sum()  # It should be very close already but just to make sure
    smol = (0 != weights) & (weights < 0.01)
    while smol.any():
        weights[smol] = 0
        weights = weights / weights.sum()
        smol = (0 != weights) & (weights < 0.01)

    probs = np.zeros(len(action_options), dtype=np.float32)
    for i, c in enumerate(candidates):
        idx = np.where(np.all(action_options == c, axis=1))[0]
        probs[idx] = weights[i]

    return probs



from typing import List, Dict, Any

import numpy as np
from rlgym.api import RewardFunction, AgentID, StateType, RewardType
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import BACK_WALL_Y, CEILING_Z


class BallTravelReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, consecutive_weight=1.0,
                 pass_weight=1.0, receive_weight=1.0,
                 giveaway_weight=-1.0, intercept_weight=1.0,
                 goal_weight=1.0,
                 distance_normalization=None,
                 do_integral=False):
        """
        Reward function based on the distance the ball travels between touches.

        :param consecutive_weight: Weight for distance covered between consecutive touches by the same player.
        :param pass_weight: Weight for distance covered by a pass to a teammate.
        :param receive_weight: Weight for distance covered by a pass received from a teammate.
        :param giveaway_weight: Weight for distance covered by a pass (giveaway) to an opponent.
        :param intercept_weight: Weight for distance covered by a pass intercepted from an opponent.
        :param goal_weight: Weight for distance covered between a touch and a goal.
        :param distance_normalization: Factor to normalize distance travelled between touches.
                                       Defaults to weighting a distance of the full length of the field as 1.0
        :param do_integral: Whether to calculate the area under the ball's travel curve instead of the distance.
        """
        self.consecutive_weight = consecutive_weight
        self.pass_weight = pass_weight
        self.receive_weight = receive_weight
        self.giveaway_weight = giveaway_weight
        self.intercept_weight = intercept_weight
        self.goal_weight = goal_weight

        if distance_normalization is None:
            if do_integral:
                # Use the area of half a field length by half ceiling height
                distance_normalization = 4 / (2 * BACK_WALL_Y * CEILING_Z)
            else:
                # Use the full length of the field
                distance_normalization = 1 / (2 * BACK_WALL_Y)
        self.normalization_factor = distance_normalization
        self.do_integral = do_integral

        self.prev_ball_pos = None
        self.last_touch_agent = None
        self.distance_since_touch = 0

    def reset(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> None:
        self.prev_ball_pos = initial_state.ball.position
        self.last_touch_agent = None
        self.distance_since_touch = 0

    def get_rewards(self, agents: List[AgentID], state: GameState, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, float]:
        ball_pos = state.ball.position

        # Update the distance travelled by the ball
        distance = np.linalg.norm(ball_pos - self.prev_ball_pos)
        if self.do_integral:
            # The path of the ball defines a right trapezoid (to a close approximation).
            z_height = (ball_pos[2] + self.prev_ball_pos[2]) / 2
            area = distance * z_height
            distance = area
        self.prev_ball_pos = ball_pos
        self.distance_since_touch += distance

        # Assign rewards based on the ball touches
        rewards = {k: 0.0 for k in agents}
        touching_agents = []  # This list is to remove dependence on agent order
        for agent in agents:
            car = state.cars[agent]
            if car.ball_touches > 0:
                if self.last_touch_agent is not None:
                    norm_dist = self.distance_since_touch * self.normalization_factor
                    if agent == self.last_touch_agent:
                        # Consecutive touches
                        rewards[agent] += norm_dist * self.consecutive_weight
                    elif car.team_num == state.cars[self.last_touch_agent].team_num:
                        # Pass to teammate
                        rewards[agent] += norm_dist * self.receive_weight
                        rewards[self.last_touch_agent] += norm_dist * self.pass_weight
                    else:
                        # Team change
                        rewards[agent] += norm_dist * self.intercept_weight
                        rewards[self.last_touch_agent] += norm_dist * self.giveaway_weight
                touching_agents.append(agent)
            elif car.is_demoed and self.last_touch_agent == agent:
                self.last_touch_agent = None

        if state.goal_scored and self.last_touch_agent is not None:
            team = state.scoring_team
            norm_dist = self.distance_since_touch * self.normalization_factor
            mul = 1 if team == state.cars[self.last_touch_agent].team_num else -1
            rewards[self.last_touch_agent] += mul * norm_dist * self.goal_weight

        if len(touching_agents) > 0:
            self.distance_since_touch = 0
            # Update the last touch agent
            if len(touching_agents) == 1:
                self.last_touch_agent = touching_agents[0]
            else:
                # If multiple agents touch the ball in the same step, adjust rewards
                for agent in agents:
                    rewards[agent] /= len(touching_agents)
                # and set last touch to be the one that is closest to the ball
                closest_agent = min(touching_agents,
                                    key=lambda x: np.linalg.norm(state.cars[x].physics.position - ball_pos))
                self.last_touch_agent = closest_agent

        shared_info["last_touch_agent"] = self.last_touch_agent
        shared_info["distance_since_touch"] = self.distance_since_touch

        return rewards



from typing import List, Dict, Any

import RocketSim as rs
import numpy as np
from rlgym.api import SharedInfoProvider, AgentID
from rlgym.rocket_league.api import GameState, PhysicsObject
from rlgym.rocket_league.common_values import TICKS_PER_SECOND


class BallPredictionProvider(SharedInfoProvider[AgentID, GameState]):
    def __init__(self, limit_seconds, step_seconds, game_mode=rs.GameMode.SOCCAR):
        self.limit_ticks = int(limit_seconds * TICKS_PER_SECOND)
        self.step_ticks = int(step_seconds * TICKS_PER_SECOND)
        assert self.limit_ticks % self.step_ticks == 0, "limit_seconds must be divisible by step_seconds"
        self.ball_predictor = rs.BallPredictor(game_mode)
        self.ball_prediction = None
        self.prev_tick_count = 0

    def create(self, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        if "ball_prediction" not in shared_info:
            shared_info["ball_prediction"] = None
        return shared_info

    def _update_prediction(self, ball: PhysicsObject, ticks_passed: int):
        rs_ball_prediction = self.ball_predictor.get_ball_prediction(
            rs.BallState(
                pos=rs.Vec(*ball.position),
                rot_mat=rs.RotMat(*ball.rotation_mtx.transpose().flatten()),
                vel=rs.Vec(*ball.linear_velocity),
                ang_vel=rs.Vec(*ball.angular_velocity)
            ),
            ticks_passed,
            self.limit_ticks // self.step_ticks,
            self.step_ticks
        )
        ball_prediction = [PhysicsObject() for _ in range(len(rs_ball_prediction))]
        for ball, prediction in zip(ball_prediction, rs_ball_prediction):
            ball.position = np.array([prediction.pos.x, prediction.pos.y, prediction.pos.z])
            ball.rotation_mtx = prediction.rot_mat.as_numpy().T
            ball.linear_velocity = np.array([prediction.vel.x, prediction.vel.y, prediction.vel.z])
            ball.angular_velocity = np.array([prediction.ang_vel.x, prediction.ang_vel.y, prediction.ang_vel.z])
            ball_prediction.append(ball)
        self.ball_prediction = ball_prediction

    def set_state(self, agents: List[AgentID], initial_state: GameState, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        self._update_prediction(initial_state.ball, 0)
        shared_info["ball_prediction"] = self.ball_prediction
        return shared_info

    def step(self, agents: List[AgentID], state: GameState, shared_info: Dict[str, Any]) -> Dict[str, Any]:
        self._update_prediction(state.ball, state.tick_count - self.prev_tick_count)
        self.prev_tick_count = state.tick_count
        shared_info["ball_prediction"] = self.ball_prediction
        return shared_info



import math

import numpy as np

from rlgym.rocket_league.api import PhysicsObject
from rlgym.rocket_league.common_values import TICKS_PER_SECOND, GRAVITY, BACK_WALL_Y, BALL_RADIUS, GOAL_CENTER_TO_POST, \
    BALL_RESTING_HEIGHT, GOAL_THRESHOLD


# Utility functions for ball physics


def solve_parabolic_trajectory(ball: PhysicsObject, g=GRAVITY):
    # Solve for the time of impact with the ground
    z = ball.position[2] - BALL_RESTING_HEIGHT
    vz = ball.linear_velocity[2]
    a = -0.5 * g
    b = vz
    c = z
    discriminant = b * b - 4 * a * c
    if discriminant < 0:
        return None, None
    discriminant = math.sqrt(discriminant)
    t_neg = (-b + discriminant) / (2 * a)
    t_pos = (-b - discriminant) / (2 * a)
    return t_neg, t_pos


def ball_hit_ground(ticks_passed: int, ball: PhysicsObject, pre=False):
    if ball.position[2] <= BALL_RESTING_HEIGHT:
        return True
    t_neg, t_pos = solve_parabolic_trajectory(ball)
    if t_neg is None or t_pos is None:
        return False
    if pre:
        # Positive solution, e.g. looking in the future
        # Less accurate since ball can hit something else before the ground.
        if 0 <= t_pos <= ticks_passed / TICKS_PER_SECOND:
            return True
    else:
        # Negative solution, e.g. looking in the past.
        # Generally preferred as it's more likely to be correct.
        if -ticks_passed / TICKS_PER_SECOND <= t_neg <= 0:
            return True
    return False


def closest_point_in_goal(ball_pos, margin=BALL_RADIUS):
    # Find the closest point on each goal to the ball
    x = math.copysign(1, ball_pos[0]) * min(abs(ball_pos[0]), GOAL_CENTER_TO_POST - margin)
    y = BACK_WALL_Y + margin
    z = min(ball_pos[2], GOAL_THRESHOLD)
    return np.array([x, y, z])



from typing import Dict, Any

import numpy as np
from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState

from rlgym_tools.rocket_league.math.skellam import AVERAGE_EPISODE_LENGTH
from rlgym_tools.rocket_league.shared_info_providers.scoreboard_provider import ScoreboardInfo


class RandomScoreboardMutator(StateMutator[GameState]):
    def __init__(self, max_game_length: float = 300.0):
        self.max_game_length = max_game_length

    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        scoreboard = shared_info.get("scoreboard", None)
        if scoreboard is None:
            scoreboard = ScoreboardInfo(300, 0, 0, 0, False, False)
            shared_info["scoreboard"] = scoreboard

        num_cars = len(state.cars)
        ael = AVERAGE_EPISODE_LENGTH[num_cars // 2 - 1]  # Average episode length for this gamemode

        # First, check if we should go to overtime
        rate = 0.5 / ael  # Rate of scoring for each team
        b = np.random.poisson(lam=self.max_game_length * rate)
        o = np.random.poisson(lam=self.max_game_length * rate)
        if b == o:
            # We've "played an entire game and ended up equal", so we go to overtime
            scoreboard.game_timer_seconds = float("inf")
        else:
            # No overtime, set to random time between 0 and max_game_length and score based on that
            time_passed = np.random.uniform(0, self.max_game_length)
            b = np.random.poisson(lam=time_passed * rate)
            o = np.random.poisson(lam=time_passed * rate)
            scoreboard.game_timer_seconds = self.max_game_length - time_passed
        scoreboard.blue_score = b
        scoreboard.orange_score = o



from dataclasses import dataclass
from typing import Dict, Optional

import numpy as np
from rlgym.rocket_league.api import GameState

from rlgym_tools.rocket_league.shared_info_providers.scoreboard_provider import ScoreboardInfo


@dataclass(slots=True)
class ReplayFrame:
    state: GameState
    actions: Dict[int, np.ndarray]
    update_age: Dict[int, float]
    scoreboard: ScoreboardInfo
    episode_seconds_remaining: float
    next_scoring_team: Optional[int]
    winning_team: Optional[int]



from setuptools import setup, find_packages

__version__ = None  # This will get replaced when reading version.py
exec(open('rlgym_tools/version.py').read())

with open('README.md', 'r') as readme_file:
    long_description = readme_file.read()

setup(
    name='rlgym_tools',
    packages=find_packages(),
    version=__version__,
    description='Extra tools for RLGym.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rolv-Arild Braaten, Lucas Emery and Matthew Allen',
    url='https://rlgym.github.io',
    install_requires=[
        'rlgym>=2.0.0',  # Including release candidates starting from 2.0.0
        'rlgym-rocket-league[sim]>=2.0.0'  # Remove if we ever add other envs
    ],
    python_requires='>=3.7',
    license='Apache 2.0',
    license_file='LICENSE',
    keywords=['rocket-league', 'gym', 'reinforcement-learning'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        "Operating System :: OS Independent",
    ],
    package_data={'rlgym_tools': ['rocket_league/replays/carball.exe']},
    include_package_data=True,  # For MANIFEST.in
)



import math
from typing import List, Dict, Any, Callable

from rlgym.api import RewardFunction, AgentID, StateType, RewardType
from rlgym.rocket_league.api import GameState


class BoostChangeReward(RewardFunction[AgentID, GameState, float]):
    def __init__(self, gain_weight: float = 1.0, lose_weight=1.0,
                 activation_fn: Callable[[float], float] = lambda x: math.sqrt(0.01 * x)):
        """
        Reward function that rewards agents for increasing their boost and penalizes them for decreasing it.

        :param gain_weight: Weight to apply to the reward when the agent gains boost
        :param lose_weight: Weight to apply to the reward when the agent loses boost
        :param activation_fn: Activation function to apply to the boost value before calculating the reward. Default is
                              the square root function so that increasing boost is more important when boost is low.
        """
        self.gain_weight = gain_weight
        self.lose_weight = lose_weight
        self.activation_fn = activation_fn

        self.prev_values = None

    def reset(self, agents: List[AgentID], initial_state: StateType, shared_info: Dict[str, Any]) -> None:
        self.prev_values = {
            agent: self.activation_fn(initial_state.cars[agent].boost_amount)
            for agent in agents
        }

    def get_rewards(self, agents: List[AgentID], state: StateType, is_terminated: Dict[AgentID, bool],
                    is_truncated: Dict[AgentID, bool], shared_info: Dict[str, Any]) -> Dict[AgentID, RewardType]:
        rewards = {}
        for agent in agents:
            current_value = self.activation_fn(state.cars[agent].boost_amount)
            delta = current_value - self.prev_values[agent]
            if delta > 0:
                rewards[agent] = delta * self.gain_weight
            elif delta < 0:
                rewards[agent] = delta * self.lose_weight
            else:
                rewards[agent] = 0
            self.prev_values[agent] = current_value

        return rewards



import random
from typing import Dict, Any

import numpy as np

from rlgym.api import StateMutator
from rlgym.rocket_league.api import GameState
from rlgym.rocket_league.common_values import BLUE_TEAM, ORANGE_TEAM, BOOST_LOCATIONS

boost_locations = np.array(BOOST_LOCATIONS)
_left_right_pad_indices = np.empty(len(boost_locations), dtype=int)
_front_back_pad_indices = np.empty(len(boost_locations), dtype=int)
for i in range(len(boost_locations)):
    _left_right_pad_indices[i] = np.argmin(
        np.linalg.norm(boost_locations - boost_locations[i] * [-1, 1, 1], axis=1))
    _front_back_pad_indices[i] = np.argmin(
        np.linalg.norm(boost_locations - boost_locations[i] * [1, -1, 1], axis=1))


class AugmentMutator(StateMutator[GameState]):
    def __init__(self, shuffle_within_teams: bool = True,
                 randomize_front_back: bool = True, randomize_left_right: bool = True):
        self.shuffle_within_teams = shuffle_within_teams
        self.randomize_front_back = randomize_front_back
        self.randomize_left_right = randomize_left_right

    def apply(self, state: GameState, shared_info: Dict[str, Any]) -> None:
        bits = random.getrandbits(2)

        scoreboard = shared_info.get("scoreboard")

        if self.shuffle_within_teams:
            self.shuffle_players(state)
        if self.randomize_front_back and bits & 1:
            self.swap_front_back(state)
            if scoreboard is not None:
                scoreboard.blue_score, scoreboard.orange_score = scoreboard.orange_score, scoreboard.blue_score
        if self.randomize_left_right and bits & 2:
            self.swap_left_right(state)

    @staticmethod
    def shuffle_players(state):
        for team in BLUE_TEAM, ORANGE_TEAM:
            agents = [agent for agent, car in state.cars.items() if car.team_num == team]
            if len(agents) == 1:
                continue
            from_agents = list(agents)
            to_agents = list(agents)
            random.shuffle(to_agents)
            mapping = dict(zip(from_agents, to_agents))
            original_cars = {k: v for k, v in state.cars.items()}
            for from_agent, to_agent in mapping.items():
                from_car = original_cars.pop(from_agent)
                if from_car.bump_victim_id is not None:
                    from_car.bump_victim_id = mapping[from_car.bump_victim_id]
                state.cars[to_agent] = from_car

    @staticmethod
    def swap_front_back(state):  # Across x-axis, meaning y-axis is inverted
        state.ball.position[1] *= -1
        state.ball.linear_velocity[1] *= -1
        state.ball.angular_velocity[0] *= -1
        state.ball.angular_velocity[2] *= -1
        for car in state.cars.values():
            car.team_num = BLUE_TEAM if car.is_orange else ORANGE_TEAM
            car.physics.position[1] *= -1
            car.physics.linear_velocity[1] *= -1
            car.physics.angular_velocity[0] *= -1
            car.physics.angular_velocity[2] *= -1
            rot_mtx = car.physics.rotation_mtx.copy()
            # Invert y-axis for fwd and up vectors, and x+z axes for right vector (to maintain handedness)
            rot_mtx[1, 0] *= -1
            rot_mtx[0, 1] *= -1
            rot_mtx[2, 1] *= -1
            rot_mtx[1, 2] *= -1
            car.physics.rotation_mtx = rot_mtx
            wwc = car.wheels_with_contact
            car.wheels_with_contact = wwc[1], wwc[0], wwc[3], wwc[0]  # Swap wheels
        state.boost_pad_timers[:] = state.boost_pad_timers[_front_back_pad_indices]

    @staticmethod
    def swap_left_right(state):  # Across y-axis, meaning x-axis is inverted
        state.ball.position[0] *= -1
        state.ball.linear_velocity[0] *= -1
        state.ball.angular_velocity[1] *= -1
        state.ball.angular_velocity[2] *= -1
        for car in state.cars.values():
            car.physics.position[0] *= -1
            car.physics.linear_velocity[0] *= -1
            car.physics.angular_velocity[1] *= -1
            car.physics.angular_velocity[2] *= -1
            rot_mtx = car.physics.rotation_mtx.copy()
            # Invert x-axis for fwd and up vectors, and y+z axes for right vector (to maintain handedness)
            rot_mtx[0, 0] *= -1
            rot_mtx[1, 1] *= -1
            rot_mtx[2, 1] *= -1
            rot_mtx[0, 2] *= -1
            car.physics.rotation_mtx = rot_mtx
            wwc = car.wheels_with_contact
            car.wheels_with_contact = wwc[1], wwc[0], wwc[3], wwc[0]  # Swap wheels
        state.boost_pad_timers[:] = state.boost_pad_timers[_left_right_pad_indices]



from dataclasses import dataclass

import numpy as np


@dataclass(slots=True)
class Action:
    """
    Dataclass representing an action in the environment.
    """
    throttle: float = 0.0
    steer: float = 0.0
    pitch: float = 0.0
    yaw: float = 0.0
    roll: float = 0.0
    jump: bool = False
    boost: bool = False
    handbrake: bool = False

    @classmethod
    def from_numpy(cls, action: np.ndarray):
        return cls(*action[:5], *action[5:].astype(bool))

    def to_numpy(self):
        return np.array(
            [self.throttle, self.steer, self.pitch, self.yaw, self.roll, self.jump, self.boost, self.handbrake],
            dtype=np.float32
        )

    def __repr__(self):
        def format_bool(b):
            return int(b)

        def format_float(f):
            return f"{f:+.2f}"

        return (f"Action("
                f"throttle={format_float(self.throttle)}, "
                f"steer={format_float(self.steer)}, "
                f"pitch={format_float(self.pitch)}, "
                f"yaw={format_float(self.yaw)}, "
                f"roll={format_float(self.roll)}, "
                f"jump={format_bool(self.jump)}, "
                f"boost={format_bool(self.boost)}, "
                f"handbrake={format_bool(self.handbrake)}"
                f")")
